#ifndef FARRAY_HPP
#define FARRAY_HPP

#include <fstream.h>
#include <stdlib.h> // size_t
#include <string.h> // strncmp

template <class T> class FileArrayElementAccess {};

template <class T> class FileArray;
template <class T> class FileArrayProxy
{
public:
  operator T() const;
  FileArrayProxy<T>& operator=(const T& t);
  FileArrayProxy<T>& operator=(const FileArrayProxy<T>& p);
private:
  FileArrayProxy(FileArray<T>& f, size_t i);

  operator T&() const;
  T* operator&(void) const;
  size_t index;
  FileArray<T>& fa;

  friend class FileArray<T>;
};

template <class T>
ostream& operator<<(ostream& os, const FileArrayProxy<T>& p)
{
  return os << p.operator T();
}

template <class T>
class FileArray
{
public:
  FileArray(fstream& fs, size_t elements); // create one.
  FileArray(fstream& fs); // use existing array
  ~FileArray(void);
  T operator[](size_t size) const;
  FileArrayProxy<T> operator[](size_t size);
  size_t size() const;
private:
  FileArray(const FileArray<T>&); // illegal
  FileArray<T>& operator=(const FileArray<T>&); // illegal

  void initFromFile(const char*);

  T readElement(size_t index) const;
  void storeElement(size_t index, const T& elem);
  
  fstream& stream;
  size_t array_size;
  streampos home;

  friend class FileArrayProxy<T>;
};

template <class T>
FileArray<T>::~FileArray()
{
}

template <class T>
FileArrayProxy<T>::FileArrayProxy(FileArray<T>& f, size_t i)
  : index(i),
    fa(f)
{
}

template <class T>
FileArrayProxy<T>::operator T() const
{
  return fa.readElement(index);
}

template <class T>
FileArrayProxy<T>& FileArrayProxy<T>::operator=(const T& t)
{
  fa.storeElement(index,t);
  return *this;
}

template <class T>
FileArrayProxy<T>& FileArrayProxy<T>::operator=(const FileArrayProxy<T>& p)
{
  fa.storeElement(index,p.fa.readElement(p.index));
  return *this;
}

template <class T>
FileArray<T>::FileArray(fstream& fs, size_t elements)
  : stream(fs),
    array_size(elements)
{
  // what if the file could not be opened?

  // first try tp read and see if there's a begin
  // pattern. Either there is one, or we should
  // get an eof.

  char pattern[6];
  stream.read(pattern,6);
  if (stream.eof()) {
    stream.clear(); // clear error state
                    // and initialize.
    
    // begin of array pattern.
    stream.write("ABegin",6);

    // must store size of elements, as last month
    const size_t elem_size
      =FileArrayElementAccess<T>::size;
    stream.write((const char*)&elem_size,
		 sizeof(elem_size));

    // and of course the number of elements
    stream.write((const char*)&array_size,
		 sizeof(array_size));

    // Now that we've written the maintenance
    // stuff, we know what the home position is.
    home = stream.tellp();

    // Then we must go the the end and write
    // the end pattern.

    stream.seekp(home+elem_size*array_size);
    stream.write("AEnd",4);

    // set put and get pointer to past the end pos.
    stream.seekg(stream.tellp());
    return;
  }
  initFromFile(pattern); // shared with other
                         // stream constructor
  if (array_size != elements) {
    // Uh oh. The data read from the stream,
    // and the size given in the constructor
    // mismatches! What now?
    stream.clear(ios::failbit);
  }
  // set put and get pointer to past the end pos.
  stream.seekp(stream.tellg());
}

template <class T>
FileArray<T>::FileArray(fstream& fs)
  : stream(fs)
{
  // First read the head pattern to see if
  // it's right.
  char pattern[6];
  stream.read(pattern,6);
  initFromFile(pattern);
  // set put and get pointer to past the end pos.
  stream.seekp(stream.tellg());
}

template <class T>
void FileArray<T>::initFromFile(const char* p)
{
  // Check if the read pattern is correct
  if (strncmp(p,"ABegin",6)) {
    // What to do? It was all wrong!
    stream.clear(ios::failbit); // for lack of better,
                                // set the fail flag.
    return;
  }
  // OK, we have a valid array, now let's see if
  // it's of the right kind.
  size_t elem_size;
  stream.read((char*)&elem_size,sizeof(elem_size));
  if (elem_size != FileArrayElementAccess<T>::size)
  {
    // wrong kind of array, the element sizes
    // mismatch. Again, what to do? Let's set
    // the fail flag for now.
    stream.clear(ios::failbit); // stupid name for the
                               // member function, right?
    return;
  }

  // Get the size of the array. Can't do much with
  // the size here, though.
  stream.read((char*)&array_size,sizeof(array_size));

  // Now we're past the header, so we know where the
  // data begins and can set the home position.

  home = stream.tellg();
  
  stream.seekg(home+elem_size*array_size);
  // Now positioned immediately after the last
  // element.
  char epattern[4];
  stream.read(epattern,4);
  if (strncmp(epattern,"AEnd",4)) {
    // Whoops, corrupt file!
    stream.clear(ios::failbit);
    return;
  }
  // Seems like we have a valid array!
}

template <class T>
T FileArray<T>::operator[](size_t size) const
{
  // what if size >= max_size?
  return readElement(size);
}

template <class T>
FileArrayProxy<T> FileArray<T>::operator[](size_t size)
{
  // what if size >= max_size?
  return FileArrayProxy<T>(*this , size);
}

template <class T>
size_t FileArray<T>::size() const
{
  return array_size;
}

template <class T>
T FileArray<T>::readElement(size_t index) const
{
  // what if index >= max_elements?
  typedef FileArrayElementAccess<T> traits;
  stream.seekg(home+index*traits::size);
  // what if seek fails?

  return traits::readFrom(stream);
  // what if read fails?
  // What if too much data is read?
}

template <class T>
void FileArray<T>::storeElement(size_t index, const T& element)
{
  // what if index >= array_size?
  typedef FileArrayElementAccess<T> traits;
  stream.seekp(home+traits::size*index);
  // what if seek fails?
  traits::writeTo(element,stream);
  // what if wite failed?
  // what if too much data was written?
}

class FileArrayElementAccess<char>
{
public:
  typedef char type;
  enum { size = sizeof(type) };
  static void writeTo(type c, ostream& os) { os.put(c); };
  static type readFrom(istream& is) { type c; is.get(c); return c; };
};

class FileArrayElementAccess<signed char>
{
public:
  typedef signed char type;
  enum { size = sizeof(type) };
  static void writeTo(type c, ostream& os) { os.put(c); };
  static type readFrom(istream& is) { type c; is.get(c); return c; };
};

class FileArrayElementAccess<unsigned char>
{
public:
  typedef unsigned char type;
  enum { size = sizeof(type) };
  static void writeTo(type c, ostream& os) { os.put(c); };
  static type readFrom(istream& is) { type c; is.get(c); return c; };
};

class FileArrayElementAccess<short>
{
public:
  typedef short type;
  enum { size = sizeof(type) };
  static void writeTo(type v, ostream& os) {os.write((const char*)&v,size);};
  static type readFrom(istream& is){type t;is.read((char*)&t,size);return t;};
};

class FileArrayElementAccess<unsigned short> 
{
public:
  typedef unsigned short type;
  enum { size = sizeof(type) };
  static void writeTo(type v, ostream& os) {os.write((const char*)&v,size);};
  static type readFrom(istream& is){type t;is.read((char*)&t,size);return t;};
};

class FileArrayElementAccess<int> 
{
public:
  typedef int type;
  enum { size = sizeof(type) };
  static void writeTo(type v, ostream& os) {os.write((const char*)&v,size);};
  static type readFrom(istream& is){type t;is.read((char*)&t,size);return t;};
};

class FileArrayElementAccess<unsigned>
{
public:
  typedef unsigned type;
  enum { size = sizeof(type) };
  static void writeTo(type v, ostream& os) {os.write((const char*)&v,size);};
  static type readFrom(istream& is){type t;is.read((char*)&t,size);return t;};
};

class FileArrayElementAccess<long>
{
public:
  typedef long type;
  enum { size = sizeof(type) };
  static void writeTo(type v, ostream& os) {os.write((const char*)&v,size);};
  static type readFrom(istream& is){type t;is.read((char*)&t,size);return t;};
};

class FileArrayElementAccess<unsigned long>
{
public:
  typedef unsigned long type;
  enum { size = sizeof(type) };
  static void writeTo(type v, ostream& os) {os.write((const char*)&v,size);};
  static type readFrom(istream& is){type t;is.read((char*)&t,size); return t;};
};

class FileArrayElementAccess<float>
{
public:
  typedef float type;
  enum { size = sizeof(type) };
  static void writeTo(type v, ostream& os) {os.write((const char*)&v,size);};
  static type readFrom(istream& is){type t;is.read((char*)&t,size);return t;};
};

class FileArrayElementAccess<double>
{
public:
  typedef double type;
  enum { size = sizeof(type) };
  static void writeTo(type v, ostream& os) {os.write((const char*)&v,size);};
  static type readFrom(istream& is){type t;is.read((char*)&t,size);return t;};
};

class FileArrayElementAccess<long double>
{
public:
  typedef long double type;
  enum { size = sizeof(type) };
  static void writeTo(type v, ostream& os) {os.write((const char*)&v,size);};
  static type readFrom(istream& is){type t;is.read((char*)&t,size);return t;};
};
  
#endif // FARRAY_HPP

